@@ -1,4 +1,4 @@
-// Copyright 2019, 2020 The Godror Authors
+// Copyright 2019, 2022 The Godror Authors
 //
 //
 // SPDX-License-Identifier: UPL-1.0 OR Apache-2.0
@@ -6,19 +6,22 @@
 package dsn
 
 import (
+	"context"
+	"database/sql"
 	"database/sql/driver"
-	"encoding/base64"
+	"encoding"
+	"errors"
 	"fmt"
-	"hash/fnv"
 	"io"
 	"net/url"
 	"sort"
 	"strconv"
 	"strings"
+	"sync"
 	"time"
 
 	"github.com/go-logfmt/logfmt"
-	errors "golang.org/x/xerrors"
+	"github.com/godror/godror/slog"
 )
 
 const (
@@ -41,28 +44,61 @@ const (
 	DefaultWaitTimeout = 30 * time.Second
 	// DefaultMaxLifeTime is the maximum time in seconds till a pooled session may exist
 	DefaultMaxLifeTime = 1 * time.Hour
-	//DefaultStandaloneConnection holds the default for standaloneConnection.
-	DefaultStandaloneConnection = false
+	// DefaultStandaloneConnection holds the default for standaloneConnection.
+	DefaultStandaloneConnection = true
+	// DefaultNoBreakOnContextCancel holds the default for noBreakOnContext
+	DefaultNoBreakOnContextCancel = false
 )
 
-// CommonParams holds the common parameters for pooled or standalone connections.
-type CommonParams struct {
-	Username, ConnectString string
-	Password                Password
+type CommonSimpleParams struct {
+	Timezone                *time.Location
 	ConfigDir, LibDir       string
+	Username, ConnectString string
+	Token, PrivateKey       string
+	Password                Password
+	Charset                 string
+	// StmtCacheSize of 0 means the default, -1 to disable the stmt cache completely
+	StmtCacheSize int
+	// true: OnInit will be called only by the new session / false: OnInit will called by new or pooled connection
+	InitOnNewConn                               bool
+	EnableEvents, NoTZCheck, PerSessionTimezone bool
+	NoBreakOnContextCancel                      bool
+}
+
+// CommonParams holds the common parameters for pooled or standalone connections.
+//
+// For details, see https://oracle.github.io/odpi/doc/structs/dpiCommonCreateParams.html#dpicommoncreateparams
+type CommonParams struct {
+	// Logger is the per-pool or per-connection logger.
+	// The default nil logger does not log.
+	Logger *slog.Logger
 	// OnInit is executed on session init. Overrides AlterSession and OnInitStmts!
-	OnInit func(driver.Conn) error
+	OnInit func(context.Context, driver.ConnPrepareContext) error `json:"-"`
 	// OnInitStmts are executed on session init, iff OnInit is nil.
 	OnInitStmts []string
 	// AlterSession key-values are set with "ALTER SESSION SET key=value" on session init, iff OnInit is nil.
 	AlterSession [][2]string
-	Timezone     *time.Location
-	EnableEvents bool
+	CommonSimpleParams
 }
 
-// String returns the string representation of CommonParams.
 func (P CommonParams) String() string {
-	q := newParamsArray(8)
+	return P.CommonSimpleParams.String()
+}
+
+var cacheCPSMu sync.RWMutex
+var cacheCPS = make(map[CommonSimpleParams]string)
+
+// String returns the string representation of CommonParams.
+func (P CommonSimpleParams) String() string {
+	cacheCPSMu.RLock()
+	s, ok := cacheCPS[P]
+	cacheCPSMu.RUnlock()
+	if ok {
+		return s
+	}
+
+	q := acquireParamsArray(8)
+	defer releaseParamsArray(q)
 	q.Add("user", P.Username)
 	q.Add("password", P.Password.String())
 	q.Add("connectString", P.ConnectString)
@@ -72,44 +108,70 @@ func (P CommonParams) String() string {
 	if P.LibDir != "" {
 		q.Add("libDir", P.LibDir)
 	}
-	s := "local"
+	s = ""
 	tz := P.Timezone
-	if tz != nil && tz != time.Local {
-		s = tz.String()
+	if tz != nil {
+		if tz == time.Local {
+			s = "local"
+		} else {
+			s = tz.String()
+		}
+	}
+	if s != "" {
+		q.Add("timezone", s)
 	}
-	q.Add("timezone", s)
 	if P.EnableEvents {
 		q.Add("enableEvents", "1")
 	}
+	if P.NoTZCheck {
+		q.Add("noTimezoneCheck", "1")
+	}
+	if P.PerSessionTimezone {
+		q.Add("perSessionTimezone", "1")
+	}
+	if P.StmtCacheSize != 0 {
+		q.Add("stmtCacheSize", strconv.Itoa(int(P.StmtCacheSize)))
+	}
+	if P.Charset != "" {
+		q.Add("charset", P.Charset)
+	}
+	if P.InitOnNewConn {
+		q.Add("initOnNewConnection", "1")
+	}
+	if P.NoBreakOnContextCancel {
+		q.Add("noBreakOnContextCancel", "1")
+	}
 
-	return q.String()
+	s = q.String()
+	cacheCPSMu.Lock()
+	cacheCPS[P] = s
+	cacheCPSMu.Unlock()
+	return s
 }
 
 // ConnParams holds the connection-specific parameters.
+//
+// For details, see https://oracle.github.io/odpi/doc/structs/dpiConnCreateParams.html#dpiconncreateparams
 type ConnParams struct {
-	NewPassword                             Password
-	ConnClass                               string
-	IsSysDBA, IsSysOper, IsSysASM, IsPrelim bool
-	ShardingKey, SuperShardingKey           []interface{}
+	NewPassword                   Password
+	ConnClass                     string
+	ShardingKey, SuperShardingKey []interface{}
+	AdminRole                     AdminRole
+	IsPrelim                      bool
 }
 
 // String returns the string representation of the ConnParams.
 func (P ConnParams) String() string {
-	q := newParamsArray(8)
+	q := acquireParamsArray(8)
+	defer releaseParamsArray(q)
 	if P.ConnClass != "" {
 		q.Add("connectionClass", P.ConnClass)
 	}
 	if !P.NewPassword.IsZero() {
 		q.Add("newPassword", P.NewPassword.String())
 	}
-	if P.IsSysDBA {
-		q.Add("sysdba", "1")
-	}
-	if P.IsSysOper {
-		q.Add("sysoper", "1")
-	}
-	if P.IsSysASM {
-		q.Add("sysasm", "1")
+	if P.AdminRole != "" {
+		q.Add("adminRole", P.AdminRole.String())
 	}
 	for _, v := range P.ShardingKey {
 		q.Add("shardingKey", fmt.Sprintf("%v", v))
@@ -120,27 +182,56 @@ func (P ConnParams) String() string {
 	return q.String()
 }
 
+// AccessToken Data for Token Authentication.
+type AccessToken struct {
+	Token      string
+	PrivateKey string
+}
+
 // PoolParams holds the configuration of the Oracle Session Pool.
+//
+// For details, see https://oracle.github.io/odpi/doc/structs/dpiPoolCreateParams.html#dpipoolcreateparams
+//
+// For details, see https://oracle.github.io/odpi/doc/structs/dpiPoolCreateParams.html#dpipoolcreateparams
 type PoolParams struct {
+	TokenCBCtx                                 context.Context                           `json:"-"`
+	TokenCB                                    func(context.Context, *AccessToken) error `json:"-"`
 	MinSessions, MaxSessions, SessionIncrement int
+	MaxSessionsPerShard                        int
 	WaitTimeout, MaxLifeTime, SessionTimeout   time.Duration
-	Heterogeneous, ExternalAuth                bool
+	PingInterval                               time.Duration
+	Heterogeneous, ExternalAuth                sql.NullBool
 }
 
 // String returns the string representation of PoolParams.
 func (P PoolParams) String() string {
-	q := newParamsArray(8)
+	q := acquireParamsArray(8)
+	defer releaseParamsArray(q)
 	q.Add("poolMinSessions", strconv.Itoa(P.MinSessions))
-	q.Add("poolMaxSessions", strconv.Itoa(P.MaxSessions))
-	q.Add("poolIncrement", strconv.Itoa(P.SessionIncrement))
-	if P.Heterogeneous {
-		q.Add("heterogeneousPool", "1")
+	if P.MaxSessions != 0 {
+		q.Add("poolMaxSessions", strconv.Itoa(P.MaxSessions))
 	}
-	q.Add("poolWaitTimeout", P.WaitTimeout.String())
-	q.Add("poolSessionMaxLifetime", P.MaxLifeTime.String())
-	q.Add("poolSessionTimeout", P.SessionTimeout.String())
-	if P.ExternalAuth {
-		q.Add("externalAuth", "1")
+	if P.MaxSessionsPerShard != 0 {
+		q.Add("poolMaxSessionsPerShard", strconv.Itoa(P.MaxSessionsPerShard))
+	}
+	q.Add("poolIncrement", strconv.Itoa(P.SessionIncrement))
+	if P.Heterogeneous.Valid {
+		q.Add("heterogeneousPool", b2s(P.Heterogeneous.Bool))
+	}
+	if P.WaitTimeout != 0 {
+		q.Add("poolWaitTimeout", P.WaitTimeout.String())
+	}
+	if P.MaxLifeTime != 0 {
+		q.Add("poolSessionMaxLifetime", P.MaxLifeTime.String())
+	}
+	if P.SessionTimeout != 0 {
+		q.Add("poolSessionTimeout", P.SessionTimeout.String())
+	}
+	if P.ExternalAuth.Valid {
+		q.Add("externalAuth", b2s(P.ExternalAuth.Bool))
+	}
+	if P.PingInterval != 0 {
+		q.Add("pingInterval", P.PingInterval.String())
 	}
 	return q.String()
 }
@@ -152,29 +243,36 @@ type ConnectionParams struct {
 	CommonParams
 	ConnParams
 	PoolParams
-	// NewPassword is used iff StandaloneConnection is true!
-	NewPassword          Password
-	StandaloneConnection bool
+	// ConnParams.NewPassword is used iff StandaloneConnection is true!
+	StandaloneConnection sql.NullBool
 }
 
 // IsStandalone returns whether the connection should be standalone, not pooled.
 func (P ConnectionParams) IsStandalone() bool {
-	return P.StandaloneConnection || P.IsSysDBA || P.IsSysOper || P.IsSysASM || P.IsPrelim
+	return P.StandaloneConnection.Valid && P.StandaloneConnection.Bool ||
+		(!P.StandaloneConnection.Valid && DefaultStandaloneConnection) ||
+		P.ConnClass == NoConnectionPoolingConnectionClass ||
+		P.AdminRole != NoRole
 }
 
 func (P *ConnectionParams) comb() {
-	P.StandaloneConnection = P.StandaloneConnection || P.ConnClass == NoConnectionPoolingConnectionClass
-	if P.IsPrelim || P.StandaloneConnection {
+	if (!P.StandaloneConnection.Valid || !P.StandaloneConnection.Bool) &&
+		P.ConnClass == NoConnectionPoolingConnectionClass {
+		P.StandaloneConnection = Bool(true)
+	}
+	if P.IsPrelim || P.StandaloneConnection.Valid && P.StandaloneConnection.Bool {
 		// Prelim: the shared memory may not exist when Oracle is shut down.
 		P.ConnClass = ""
-		P.Heterogeneous = false
+		P.Heterogeneous = Bool(false)
 	}
 	if !P.IsStandalone() {
 		P.NewPassword.Reset()
 		// only enable external authentication if we are dealing with a
 		// homogeneous pool and no user name/password has been specified
-		if P.Username == "" && P.Password.IsZero() && !P.Heterogeneous {
-			P.ExternalAuth = true
+		if P.Username == "" && P.Password.IsZero() &&
+			!P.ExternalAuth.Valid &&
+			!(P.Heterogeneous.Valid && P.Heterogeneous.Bool) {
+			P.ExternalAuth = Bool(true)
 		}
 	}
 }
@@ -185,13 +283,13 @@ func (P *ConnectionParams) SetSessionParamOnInit(k, v string) {
 }
 
 // String returns the string representation of ConnectionParams.
-// The password is replaced with a "SECRET" string!
+// The password is replaced with a "***" string!
 func (P ConnectionParams) String() string {
 	return P.string(true, false)
 }
 
 // StringNoClass returns the string representation of ConnectionParams, without class info.
-// The password is replaced with a "SECRET" string!
+// The password is replaced with a "***" string!
 func (P ConnectionParams) StringNoClass() string {
 	return P.string(false, false)
 }
@@ -203,7 +301,8 @@ func (P ConnectionParams) StringWithPassword() string {
 }
 
 func (P ConnectionParams) string(class, withPassword bool) string {
-	q := newParamsArray(32)
+	q := acquireParamsArray(32)
+	defer releaseParamsArray(q)
 	q.Add("connectString", P.ConnectString)
 	s := P.ConnClass
 	if !class {
@@ -215,58 +314,111 @@ func (P ConnectionParams) string(class, withPassword bool) string {
 	if withPassword {
 		q.Add("password", P.Password.Secret())
 		q.Add("newPassword", P.NewPassword.Secret())
+		if P.Token != "" {
+			q.Add("token", P.Token)
+		}
+		if P.PrivateKey != "" {
+			q.Add("privateKey", P.PrivateKey)
+		}
 	} else {
 		q.Add("password", P.Password.String())
 		if !P.NewPassword.IsZero() {
 			q.Add("newPassword", P.NewPassword.String())
 		}
 	}
-	s = "local"
-	tz := P.Timezone
-	if tz != nil && tz != time.Local {
-		s = tz.String()
-	}
-	q.Add("timezone", s)
-	B := func(b bool) string {
-		if b {
-			return "1"
+	s = ""
+	if tz := P.Timezone; tz != nil {
+		if tz == time.Local {
+			s = "local"
+		} else {
+			s = tz.String()
 		}
-		return "0"
+	}
+	if s != "" {
+		q.Add("timezone", s)
+	}
+	if P.NoTZCheck {
+		q.Add("noTimezoneCheck", "1")
+	}
+	if P.PerSessionTimezone {
+		q.Add("perSessionTimezone", "1")
+	}
+	if P.StmtCacheSize != 0 {
+		q.Add("stmtCacheSize", strconv.Itoa(int(P.StmtCacheSize)))
+	}
+	if P.Charset != "" {
+		q.Add("charset", P.Charset)
 	}
 	q.Add("poolMinSessions", strconv.Itoa(P.MinSessions))
-	q.Add("poolMaxSessions", strconv.Itoa(P.MaxSessions))
+	if P.MaxSessions != 0 {
+		q.Add("poolMaxSessions", strconv.Itoa(P.MaxSessions))
+	}
+	if P.MaxSessionsPerShard != 0 {
+		q.Add("poolMasSessionsPerShard", strconv.Itoa(P.MaxSessionsPerShard))
+	}
 	q.Add("poolIncrement", strconv.Itoa(P.SessionIncrement))
-	q.Add("sysdba", B(P.IsSysDBA))
-	q.Add("sysoper", B(P.IsSysOper))
-	q.Add("sysasm", B(P.IsSysASM))
-	q.Add("standaloneConnection", B(P.StandaloneConnection))
-	q.Add("enableEvents", B(P.EnableEvents))
-	q.Add("heterogeneousPool", B(P.Heterogeneous))
-	q.Add("prelim", B(P.IsPrelim))
-	q.Add("poolWaitTimeout", P.WaitTimeout.String())
-	q.Add("poolSessionMaxLifetime", P.MaxLifeTime.String())
-	q.Add("poolSessionTimeout", P.SessionTimeout.String())
-	as := newParamsArray(1)
+	if P.AdminRole != "" {
+		q.Add("adminRole", P.AdminRole.String())
+	}
+	if P.StandaloneConnection.Valid {
+		q.Add("standaloneConnection", b2s(P.StandaloneConnection.Bool))
+	}
+	if P.EnableEvents {
+		q.Add("enableEvents", "1")
+	}
+	if P.Heterogeneous.Valid {
+		q.Add("heterogeneousPool", b2s(P.Heterogeneous.Bool))
+	}
+	if P.ExternalAuth.Valid {
+		q.Add("externalAuth", b2s(P.ExternalAuth.Bool))
+	}
+	if P.IsPrelim {
+		q.Add("prelim", "1")
+	}
+	if P.WaitTimeout != 0 {
+		q.Add("poolWaitTimeout", P.WaitTimeout.String())
+	}
+	if P.MaxLifeTime != 0 {
+		q.Add("poolSessionMaxLifetime", P.MaxLifeTime.String())
+	}
+	if P.SessionTimeout != 0 {
+		q.Add("poolSessionTimeout", P.SessionTimeout.String())
+	}
+	if P.PingInterval != 0 {
+		q.Add("pingInterval", P.PingInterval.String())
+	}
+	as := acquireParamsArray(1)
+	defer releaseParamsArray(as)
 	for _, kv := range P.AlterSession {
 		as.Reset()
 		as.Add(kv[0], kv[1])
 		q.Add("alterSession", strings.TrimSpace(as.String()))
 	}
+	if P.InitOnNewConn {
+		q.Add("initOnNewConnection", "1")
+	}
+	if P.NoBreakOnContextCancel {
+		q.Add("noBreakOnContextCancel", "1")
+	}
 	q.Values["onInit"] = P.OnInitStmts
-	q.Add("configDir", P.ConfigDir)
-	q.Add("libDir", P.LibDir)
+	if P.ConfigDir != "" {
+		q.Add("configDir", P.ConfigDir)
+	}
+	if P.LibDir != "" {
+		q.Add("libDir", P.LibDir)
+	}
 	//return quoteRunes(P.Username, "/@") + "/" + quoteRunes(password, "@") + "@" + P.CommonParams.ConnectString + "\n" + q.String()
 
 	return q.String()
 }
 
 // Parse parses the given connection string into a struct.
+//
+// For examples, see [../doc/connection.md](../doc/connection.md)
 func Parse(dataSourceName string) (ConnectionParams, error) {
 	P := ConnectionParams{
-		StandaloneConnection: DefaultStandaloneConnection,
-		CommonParams: CommonParams{
-			Timezone: time.Local,
-		},
+		// StandaloneConnection: DefaultStandaloneConnection,
+		//CommonParams: CommonParams{ Timezone: time.Local, },
 		ConnParams: ConnParams{
 			ConnClass: DefaultConnectionClass,
 		},
@@ -289,11 +441,12 @@ func Parse(dataSourceName string) (ConnectionParams, error) {
 		// URL
 		u, err := url.Parse(dataSourceName)
 		if err != nil {
-			return P, errors.Errorf("%s: %w", dataSourceName, err)
+			return P, fmt.Errorf("%s: %w", dataSourceName, err)
 		}
 		if usr := u.User; usr != nil {
 			P.Username = usr.Username()
-			P.Password.secret, _ = usr.Password()
+			passw, _ := usr.Password()
+			P.Password.Set(passw)
 		}
 		P.ConnectString = u.Hostname()
 		// IPv6 literal address brackets are removed by u.Hostname,
@@ -309,24 +462,26 @@ func Parse(dataSourceName string) (ConnectionParams, error) {
 		}
 		//fmt.Printf("URL=%s cs=%q host=%q port=%q path=%q\n", u, P.ConnectString, u.Host, u.Port(), u.Path)
 		q = u.Query()
-	} else if strings.Contains(dataSourceName, "\n") || // multi-line, or
+	} else if strings.IndexByte(dataSourceName, '\n') >= 0 || // multi-line, or
 		strings.Contains(dataSourceName, "connectString=") { // contains connectString
 		// This should be a proper logfmt-encoded parameter string, with connectString
 		paramsString, dataSourceName = dataSourceName, ""
 	} else {
 		// Not URL, not logfmt-ed - an old styled DSN
 		// Old, or Easy Connect, or anything
-		P.Username, P.Password.secret, dataSourceName = parseUserPassw(dataSourceName)
+		var passw string
+		P.Username, passw, dataSourceName = parseUserPassw(dataSourceName)
+		P.Password.Set(passw)
 		//fmt.Printf("dsn=%q\n", dataSourceName)
 		uSid := strings.ToUpper(dataSourceName)
 		//fmt.Printf("dataSourceName=%q SID=%q\n", dataSourceName, uSid)
 		if strings.Contains(uSid, " AS ") {
-			if P.IsSysDBA = strings.HasSuffix(uSid, " AS SYSDBA"); P.IsSysDBA {
-				dataSourceName = dataSourceName[:len(dataSourceName)-10]
-			} else if P.IsSysOper = strings.HasSuffix(uSid, " AS SYSOPER"); P.IsSysOper {
-				dataSourceName = dataSourceName[:len(dataSourceName)-11]
-			} else if P.IsSysASM = strings.HasSuffix(uSid, " AS SYSASM"); P.IsSysASM {
-				dataSourceName = dataSourceName[:len(dataSourceName)-10]
+			for _, role := range adminRoles {
+				if s := role.String(); strings.HasSuffix(uSid, " AS "+s) {
+					P.AdminRole = role
+					dataSourceName = dataSourceName[:len(dataSourceName)-len(s)]
+					break
+				}
 			}
 		}
 		P.ConnectString = dataSourceName
@@ -336,7 +491,9 @@ func Parse(dataSourceName string) (ConnectionParams, error) {
 
 	if paramsString != "" {
 		if q == nil {
-			q = make(url.Values, 32)
+			pa := acquireParamsArray(32)
+			defer releaseParamsArray(pa)
+			q = pa.Values
 		}
 		// Parse the logfmt-formatted parameters string
 		d := logfmt.NewDecoder(strings.NewReader(paramsString))
@@ -348,7 +505,13 @@ func Parse(dataSourceName string) (ConnectionParams, error) {
 				case "user":
 					P.Username = value
 				case "password":
-					P.Password.secret = value
+					P.Password.Set(value)
+				case "charset":
+					P.Charset = value
+				case "token":
+					P.Token = value
+				case "privateKey":
+					P.PrivateKey = value
 				case "alterSession", "onInit", "shardingKey", "superShardingKey":
 					q.Add(key, value)
 				default:
@@ -356,8 +519,9 @@ func Parse(dataSourceName string) (ConnectionParams, error) {
 				}
 			}
 		}
+		// fmt.Printf("Parse q=%#v\n", q)
 		if err := d.Err(); err != nil {
-			return P, errors.Errorf("parsing parameters %q: %w", paramsString, err)
+			return P, fmt.Errorf("parsing parameters %q: %w", paramsString, err)
 		}
 	}
 	//fmt.Printf("cs0=%q\n", P.ConnectString)
@@ -367,55 +531,104 @@ func Parse(dataSourceName string) (ConnectionParams, error) {
 	if vv, ok := q["connectionClass"]; ok {
 		P.ConnClass = vv[0]
 	}
-	for _, task := range []struct {
+	var sysDBA, sysOper, sysASM bool
+	boolTasks := []struct {
 		Dest *bool
 		Key  string
 	}{
-		{&P.IsSysDBA, "sysdba"},
-		{&P.IsSysOper, "sysoper"},
-		{&P.IsSysASM, "sysasm"},
+		{&sysDBA, "sysdba"},
+		{&sysOper, "sysoper"},
+		{&sysASM, "sysasm"},
 		{&P.IsPrelim, "prelim"},
 
 		{&P.EnableEvents, "enableEvents"},
-		{&P.Heterogeneous, "heterogeneousPool"},
-		{&P.StandaloneConnection, "standaloneConnection"},
-	} {
+
+		{&P.NoTZCheck, "noTimezoneCheck"},
+		{&P.PerSessionTimezone, "perSessionTimezone"},
+		{&P.InitOnNewConn, "initOnNewConnection"},
+		{&P.NoBreakOnContextCancel, "noBreakOnContextCancel"},
+	}
+	if ar := q.Get("adminRole"); len(ar) > 3 && strings.EqualFold(ar[:3], "SYS") {
+		P.AdminRole = AdminRole(strings.ToUpper(ar))
+		boolTasks = boolTasks[3:] // skip parsing
+	}
+	for _, task := range boolTasks {
 		s := q.Get(task.Key)
 		if s == "" {
 			continue
 		}
 		var err error
 		if *task.Dest, err = strconv.ParseBool(s); err != nil {
-			return P, errors.Errorf("%s=%q: %w", task.Key, s, err)
-		}
-		if task.Key == "heterogeneousPool" {
-			P.StandaloneConnection = !P.Heterogeneous
+			return P, fmt.Errorf("%s=%q: %w", task.Key, s, err)
 		}
 	}
+	if P.AdminRole == "" {
+		if sysDBA {
+			P.AdminRole = SysDBA
+		} else if sysOper {
+			P.AdminRole = SysOPER
+		} else if sysASM {
+			P.AdminRole = SysASM
+		}
+	}
+
+	for _, task := range []struct {
+		Dest *sql.NullBool
+		Key  string
+	}{
+		{&P.Heterogeneous, "heterogeneousPool"},
+		{&P.ExternalAuth, "externalAuth"},
+		{&P.StandaloneConnection, "standaloneConnection"},
+	} {
+		s := q.Get(task.Key)
+		if s == "" {
+			continue
+		}
+		b, err := strconv.ParseBool(s)
+		if err != nil {
+			return P, fmt.Errorf("%s=%q: %w", task.Key, s, err)
+		}
+		*task.Dest = Bool(b)
+	}
+	if !P.StandaloneConnection.Valid && P.Heterogeneous.Valid {
+		P.StandaloneConnection = Bool(!P.Heterogeneous.Bool)
+	}
+	// fmt.Println("parse", P.StringWithPassword())
 
 	if tz := q.Get("timezone"); tz != "" {
-		var err error
 		if strings.EqualFold(tz, "local") {
-			// P.Timezone = time.Local // already set
+			P.Timezone = time.Local
 		} else if strings.Contains(tz, "/") {
-			if P.Timezone, err = time.LoadLocation(tz); err != nil {
-				return P, errors.Errorf("%s: %w", tz, err)
+			ptz, err := time.LoadLocation(tz)
+			if err != nil {
+				return P, fmt.Errorf("%s: %w", tz, err)
 			}
+			P.Timezone = ptz
 		} else if off, err := ParseTZ(tz); err == nil {
-			P.Timezone = time.FixedZone(tz, off)
+			if off == 0 {
+				P.Timezone = time.UTC
+			} else {
+				P.Timezone = time.FixedZone(tz, off)
+			}
 		} else {
-			return P, errors.Errorf("%s: %w", tz, err)
+			return P, fmt.Errorf("%s: %w", tz, err)
 		}
+		if P.Timezone == nil {
+			P.Timezone = time.UTC
+		}
+		//} else if P.Timezone == nil {
+		//P.Timezone = time.Local
 	}
-
 	for _, task := range []struct {
 		Dest *int
 		Key  string
 	}{
 		{&P.MinSessions, "poolMinSessions"},
 		{&P.MaxSessions, "poolMaxSessions"},
+		{&P.MaxSessionsPerShard, "poolMasSessionsPerShard"},
 		{&P.SessionIncrement, "poolIncrement"},
 		{&P.SessionIncrement, "sessionIncrement"},
+		{&P.StmtCacheSize, "stmtCacheSize"},
 	} {
 		s := q.Get(task.Key)
 		if s == "" {
@@ -424,9 +637,10 @@ func Parse(dataSourceName string) (ConnectionParams, error) {
 		var err error
 		*task.Dest, err = strconv.Atoi(s)
 		if err != nil {
-			return P, errors.Errorf("%s: %w", task.Key+"="+s, err)
+			return P, fmt.Errorf("%s: %w", task.Key+"="+s, err)
 		}
 	}
+
 	for _, task := range []struct {
 		Dest *time.Duration
 		Key  string
@@ -434,6 +648,7 @@ func Parse(dataSourceName string) (ConnectionParams, error) {
 		{&P.SessionTimeout, "poolSessionTimeout"},
 		{&P.WaitTimeout, "poolWaitTimeout"},
 		{&P.MaxLifeTime, "poolSessionMaxLifetime"},
+		{&P.PingInterval, "pingInterval"},
 	} {
 		s := q.Get(task.Key)
 		if s == "" {
@@ -443,11 +658,11 @@ func Parse(dataSourceName string) (ConnectionParams, error) {
 		*task.Dest, err = time.ParseDuration(s)
 		if err != nil {
 			if !strings.Contains(err.Error(), "time: missing unit in duration") {
-				return P, errors.Errorf("%s: %w", task.Key+"="+s, err)
+				return P, fmt.Errorf("%s: %w", task.Key+"="+s, err)
 			}
 			i, err := strconv.Atoi(s)
 			if err != nil {
-				return P, errors.Errorf("%s: %w", task.Key+"="+s, err)
+				return P, fmt.Errorf("%s: %w", task.Key+"="+s, err)
 			}
 			base := time.Second
 			if task.Key == "poolWaitTimeout" {
@@ -472,21 +687,21 @@ func Parse(dataSourceName string) (ConnectionParams, error) {
 			}
 		}
 		if err := d.Err(); err != nil {
-			return P, errors.Errorf("%q: %w", s, err)
+			return P, fmt.Errorf("%q: %w", s, err)
 		}
 	}
 	P.OnInitStmts = q["onInit"]
 	P.ShardingKey = strToIntf(q["shardingKey"])
 	P.SuperShardingKey = strToIntf(q["superShardingKey"])
 
-	P.NewPassword.secret = q.Get("newPassword")
+	P.NewPassword.Set(q.Get("newPassword"))
 	P.ConfigDir = q.Get("configDir")
 	P.LibDir = q.Get("libDir")
 
-	//fmt.Printf("cs1=%q\n", P.ConnectString)
+	// fmt.Printf("cs1=%q\n", P)
 	P.comb()
 
-	//fmt.Printf("cs2=%q\n", P.ConnectString)
+	// fmt.Printf("cs2=%q\n", P)
 
 	return P, nil
 }
@@ -497,17 +712,16 @@ type Password struct {
 }
 
 // NewPassword creates a new Password, containing the given secret.
-func NewPassword(secret string) Password { return Password{secret: secret} }
+func NewPassword(secret string) Password {
+	var P Password
+	P.Set(secret)
+	return P
+}
+
+const obfuscatedPassword = "SECRET-***"
 
 // String returns the secret obfuscated irreversibly.
-func (P Password) String() string {
-	if P.secret == "" {
-		return ""
-	}
-	hsh := fnv.New64()
-	io.WriteString(hsh, P.secret)
-	return "SECRET-" + base64.URLEncoding.EncodeToString(hsh.Sum(nil))
-}
+func (P Password) String() string { return obfuscatedPassword }
 
 // Secret reveals the real password.
 func (P Password) Secret() string { return P.secret }
@@ -522,10 +736,26 @@ func (P Password) Len() int { return len(P.secret) }
 func (P *Password) Reset() { P.secret = "" }
 
 // Set the password.
-func (P *Password) Set(secret string) { P.secret = secret }
+func (P *Password) Set(secret string) {
+	P.secret = secret
+}
+
+// LogValue implements slog.LogValuer.
+func (P Password) LogValue() slog.Value { return slog.StringValue(obfuscatedPassword) }
+
+var ErrCannotMarshal = errors.New("cannot be marshaled")
+
+func (P *Password) MarshalText() ([]byte, error) { return nil, ErrCannotMarshal }
+
+// nosemgrep
+func (P *Password) MarshalJSON() ([]byte, error)   { return nil, ErrCannotMarshal }
+func (P *Password) MarshalBinary() ([]byte, error) { return nil, ErrCannotMarshal }
+
+var _ encoding.TextMarshaler = ((*Password)(nil))
+var _ encoding.BinaryMarshaler = ((*Password)(nil))
 
 // CopyFrom another password.
-func (P *Password) CopyFrom(Q Password) { P.secret = Q.secret }
+func (P *Password) CopyFrom(Q Password) { *P = Q }
 
 // ParamsArray is an url.Values for holding parameters,
 // and logfmt-formatting them with the String() method.
@@ -533,15 +763,36 @@ type paramsArray struct {
 	url.Values
 }
 
+var paramsArrayPool8 = sync.Pool{New: func() any { return newParamsArray(8) }}
+var paramsArrayPool32 = sync.Pool{New: func() any { return newParamsArray(32) }}
+
+func acquireParamsArray(cap int) *paramsArray {
+	if cap <= 8 {
+		return paramsArrayPool8.Get().(*paramsArray)
+	}
+	return paramsArrayPool32.Get().(*paramsArray)
+}
+func releaseParamsArray(p *paramsArray) {
+	length := len(p.Values)
+	for k := range p.Values {
+		delete(p.Values, k)
+	}
+	if length < 32 {
+		paramsArrayPool8.Put(p)
+	} else {
+		paramsArrayPool32.Put(p)
+	}
+}
+
 // newParamsArray returns a new paramsArray with the given capacity of parameters.
 //
 // You can use this to build a dataSourceName for godror.
-func newParamsArray(cap int) paramsArray { return paramsArray{Values: make(url.Values, cap)} }
+func newParamsArray(cap int) *paramsArray { return &paramsArray{Values: make(url.Values, cap)} }
 
 // WriteTo the given writer, logfmt-encoded,
 // starting with username, password, connectString,
 // then the rest sorted alphabetically.
-func (p paramsArray) WriteTo(w io.Writer) (int64, error) {
+func (p *paramsArray) WriteTo(w io.Writer) (int64, error) {
 	firstKeys := make([]string, 0, len(p.Values))
 	keys := make([]string, 0, len(p.Values))
 	for k := range p.Values {
@@ -586,10 +837,13 @@ func (p paramsArray) WriteTo(w io.Writer) (int64, error) {
 	return cw.N, firstErr
 }
 
+var paSB = sync.Pool{New: func() any { return new(strings.Builder) }}
+
 // String returns the values in the params array, logfmt-formatted,
 // starting with username, password, connectString, then the rest sorted alphabetically.
-func (p paramsArray) String() string {
-	var buf strings.Builder
+func (p *paramsArray) String() string {
+	buf := paSB.Get().(*strings.Builder)
+	defer func() { buf.Reset(); paSB.Put(buf) }()
 	var n int
 	for k, vv := range p.Values {
 		for _, v := range vv {
@@ -597,39 +851,26 @@ func (p paramsArray) String() string {
 		}
 	}
 	buf.Grow(n)
-	if _, err := p.WriteTo(&buf); err != nil {
-		fmt.Fprintf(&buf, "\tERROR: %+v", err)
+	if _, err := p.WriteTo(buf); err != nil {
+		fmt.Fprintf(buf, "\tERROR: %+v", err)
 	}
 	return buf.String()
 }
-func (p paramsArray) Reset() {
+func (p *paramsArray) Reset() {
 	for k := range p.Values {
 		delete(p.Values, k)
 	}
 }
 
-/*
-func quoteRunes(s, runes string) string {
-	if !strings.ContainsAny(s, runes) {
-		return s
-	}
-	var buf strings.Builder
-	buf.Grow(2 * len(s))
-	for _, r := range s {
-		if strings.ContainsRune(runes, r) {
-			buf.WriteByte('\\')
-		}
-		buf.WriteRune(r)
-	}
-	return buf.String()
-}
-*/
+var unquoteSB = sync.Pool{New: func() any { return new(strings.Builder) }}
+
 // unquote replaces quoted ("\\n") with the quoted.
 func unquote(s string) string {
 	if !strings.ContainsRune(s, '\\') {
 		return s
 	}
-	var buf strings.Builder
+	buf := unquoteSB.Get().(*strings.Builder)
+	defer func() { buf.Reset(); unquoteSB.Put(buf) }()
 	buf.Grow(len(s))
 	var quoted bool
 	for _, r := range s {
@@ -685,20 +926,27 @@ func parseUserPassw(dataSourceName string) (user, passw, connectString string) {
 		return "", "", dataSourceName
 	}
 	ups := splitQuoted(dataSourceName, '@')
-	//if len(ups) == 1 {  // user/pass, no '@'
+	var extra string
+	if len(ups) == 1 { // user/pass, no '@'
+		if i := strings.Index(strings.ToUpper(ups[0]), " AS SYS"); i >= 0 {
+			ups[0], extra = ups[0][:i], ups[0][i:]
+		}
+	}
+	//fmt.Printf("ups=%v extra=%q\n", ups, extra)
 	userpass := splitQuoted(ups[0], '/')
 	//fmt.Printf("ups=%q\nuserpass=%q\n", ups, userpass)
 	if len(ups) == 1 && len(userpass) == 1 {
-		return "", "", dataSourceName
+		return "", "", dataSourceName + unquote(extra)
 	}
+
 	user = unquote(userpass[0])
 	if len(userpass) > 1 {
 		passw = unquote(userpass[1])
 	}
 	if len(ups) == 1 {
-		return user, passw, ""
+		return user, passw, unquote(extra)
 	}
-	return user, passw, unquote(ups[1])
+	return user, passw, unquote(ups[1] + extra)
 }
 
 // ParseTZ parses timezone specification ("Europe/Budapest" or "+01:00") and returns the offset in seconds.
@@ -713,11 +961,11 @@ func ParseTZ(s string) (int, error) {
 	var tz int
 	var ok bool
 	if i := strings.IndexByte(s, ':'); i >= 0 {
-		i64, err := strconv.ParseInt(s[i+1:], 10, 6)
+		u64, err := strconv.ParseUint(s[i+1:], 10, 6)
 		if err != nil {
-			return tz, errors.Errorf("%s: %w", s, err)
+			return tz, fmt.Errorf("%s: %w", s, err)
 		}
-		tz = int(i64 * 60)
+		tz = int(u64 * 60)
 		s = s[:i]
 		ok = true
 	}
@@ -725,7 +973,10 @@ func ParseTZ(s string) (int, error) {
 		if i := strings.IndexByte(s, '/'); i >= 0 {
 			targetLoc, err := time.LoadLocation(s)
 			if err != nil {
-				return tz, errors.Errorf("%s: %w", s, err)
+				return tz, fmt.Errorf("%s: %w", s, err)
+			}
+			if targetLoc == nil {
+				targetLoc = time.UTC
 			}
 
 			_, localOffset := time.Now().In(targetLoc).Zone()
@@ -736,7 +987,7 @@ func ParseTZ(s string) (int, error) {
 	}
 	i64, err := strconv.ParseInt(s, 10, 5)
 	if err != nil {
-		return tz, errors.Errorf("%s: %w", s, err)
+		return tz, fmt.Errorf("%s: %w", s, err)
 	}
 	if i64 < 0 {
 		tz = -tz
@@ -777,3 +1028,32 @@ func (cw *countingWriter) Write(p []byte) (int, error) {
 	cw.N += int64(n)
 	return n, err
 }
+
+func b2s(b bool) string {
+	if b {
+		return "1"
+	}
+	return "0"
+}
+
+// Bool is an sql.NullBool helper
+func Bool(b bool) sql.NullBool { return sql.NullBool{Valid: true, Bool: b} }
+
+type AdminRole string
+
+func (r AdminRole) String() string { return string(r) }
+
+const (
+	NoRole    = AdminRole("")
+	SysDBA    = AdminRole("SYSDBA")
+	SysOPER   = AdminRole("SYSOPER")
+	SysBACKUP = AdminRole("SYSBACKUP")
+	SysDG     = AdminRole("SYSDG")
+	SysKM     = AdminRole("SYSKM")
+	SysRAC    = AdminRole("SYSRAC")
+	SysASM    = AdminRole("SYSASM")
+)
+
+var adminRoles = []AdminRole{
+	SysDBA, SysOPER, SysBACKUP, SysDG, SysKM, SysRAC, SysASM,
+}
