@@ -101,8 +101,7 @@ func TestDialerRequest(t *testing.T) {
 				t.Errorf("unexpected request:\nact:\n%s\nexp:\n%s\n", act, exp)
 			}
 			if _, err := http.ReadRequest(bufio.NewReader(&buf)); err != nil {
-				t.Errorf("read request error: %s", err)
-				return
+				t.Fatalf("read request error: %s", err)
 			}
 		})
 	}
@@ -116,7 +115,7 @@ func makeAccept(nonce []byte) []byte {
 
 func BenchmarkPutAccept(b *testing.B) {
 	nonce := make([]byte, nonceSize)
-	_, err := rand.Read(nonce[:])
+	_, err := rand.Read(nonce)
 	if err != nil {
 		b.Fatal(err)
 	}
@@ -129,7 +128,7 @@ func BenchmarkPutAccept(b *testing.B) {
 
 func BenchmarkCheckNonce(b *testing.B) {
 	nonce := make([]byte, nonceSize)
-	_, err := rand.Read(nonce[:])
+	_, err := rand.Read(nonce)
 	if err != nil {
 		b.Fatal(err)
 	}
@@ -159,7 +158,7 @@ func TestDialerHandshake(t *testing.T) {
 	}{
 		{
 			res: &http.Response{
-				StatusCode: 101,
+				StatusCode: http.StatusSwitchingProtocols,
 				ProtoMajor: 1,
 				ProtoMinor: 1,
 				Header: http.Header{
@@ -174,7 +173,7 @@ func TestDialerHandshake(t *testing.T) {
 				Protocols: []string{"xml", "json", "soap"},
 			},
 			res: &http.Response{
-				StatusCode: 101,
+				StatusCode: http.StatusSwitchingProtocols,
 				ProtoMajor: 1,
 				ProtoMinor: 1,
 				Header: http.Header{
@@ -190,7 +189,7 @@ func TestDialerHandshake(t *testing.T) {
 				Protocols: []string{"xml", "json", "soap"},
 			},
 			res: &http.Response{
-				StatusCode: 101,
+				StatusCode: http.StatusSwitchingProtocols,
 				ProtoMajor: 1,
 				ProtoMinor: 1,
 				Header: http.Header{
@@ -210,7 +209,7 @@ func TestDialerHandshake(t *testing.T) {
 				},
 			},
 			res: &http.Response{
-				StatusCode: 101,
+				StatusCode: http.StatusSwitchingProtocols,
 				ProtoMajor: 1,
 				ProtoMinor: 1,
 				Header: http.Header{
@@ -231,7 +230,7 @@ func TestDialerHandshake(t *testing.T) {
 				},
 			},
 			res: &http.Response{
-				StatusCode: 101,
+				StatusCode: http.StatusSwitchingProtocols,
 				ProtoMajor: 1,
 				ProtoMinor: 1,
 				Header: http.Header{
@@ -252,7 +251,7 @@ func TestDialerHandshake(t *testing.T) {
 				},
 			},
 			res: &http.Response{
-				StatusCode: 101,
+				StatusCode: http.StatusSwitchingProtocols,
 				ProtoMajor: 1,
 				ProtoMinor: 1,
 				Header: http.Header{
@@ -267,7 +266,7 @@ func TestDialerHandshake(t *testing.T) {
 		{
 			name: "resp with frames",
 			res: &http.Response{
-				StatusCode: 101,
+				StatusCode: http.StatusSwitchingProtocols,
 				ProtoMajor: 1,
 				ProtoMinor: 1,
 				Header: http.Header{
@@ -284,7 +283,7 @@ func TestDialerHandshake(t *testing.T) {
 		{
 			name: "resp with body",
 			res: &http.Response{
-				StatusCode: 101,
+				StatusCode: http.StatusSwitchingProtocols,
 				ProtoMajor: 1,
 				ProtoMinor: 1,
 				Header: http.Header{
@@ -303,7 +302,7 @@ func TestDialerHandshake(t *testing.T) {
 		{
 			name: "bad proto",
 			res: &http.Response{
-				StatusCode: 101,
+				StatusCode: http.StatusSwitchingProtocols,
 				ProtoMajor: 2,
 				ProtoMinor: 1,
 				Header:     make(http.Header),
@@ -313,7 +312,7 @@ func TestDialerHandshake(t *testing.T) {
 		{
 			name: "bad status",
 			res: &http.Response{
-				StatusCode: 400,
+				StatusCode: http.StatusBadRequest,
 				ProtoMajor: 1,
 				ProtoMinor: 1,
 				Header:     make(http.Header),
@@ -324,7 +323,7 @@ func TestDialerHandshake(t *testing.T) {
 		{
 			name: "bad status with body",
 			res: &http.Response{
-				StatusCode: 400,
+				StatusCode: http.StatusBadRequest,
 				ProtoMajor: 1,
 				ProtoMinor: 1,
 				Header:     make(http.Header),
@@ -339,7 +338,7 @@ func TestDialerHandshake(t *testing.T) {
 		{
 			name: "bad upgrade",
 			res: &http.Response{
-				StatusCode: 101,
+				StatusCode: http.StatusSwitchingProtocols,
 				ProtoMajor: 1,
 				ProtoMinor: 1,
 				Header: http.Header{
@@ -352,7 +351,7 @@ func TestDialerHandshake(t *testing.T) {
 		{
 			name: "bad upgrade",
 			res: &http.Response{
-				StatusCode: 101,
+				StatusCode: http.StatusSwitchingProtocols,
 				ProtoMajor: 1,
 				ProtoMinor: 1,
 				Header: http.Header{
@@ -366,7 +365,7 @@ func TestDialerHandshake(t *testing.T) {
 		{
 			name: "bad connection",
 			res: &http.Response{
-				StatusCode: 101,
+				StatusCode: http.StatusSwitchingProtocols,
 				ProtoMajor: 1,
 				ProtoMinor: 1,
 				Header: http.Header{
@@ -379,7 +378,7 @@ func TestDialerHandshake(t *testing.T) {
 		{
 			name: "bad connection",
 			res: &http.Response{
-				StatusCode: 101,
+				StatusCode: http.StatusSwitchingProtocols,
 				ProtoMajor: 1,
 				ProtoMinor: 1,
 				Header: http.Header{
@@ -393,7 +392,7 @@ func TestDialerHandshake(t *testing.T) {
 		{
 			name: "bad accept",
 			res: &http.Response{
-				StatusCode: 101,
+				StatusCode: http.StatusSwitchingProtocols,
 				ProtoMajor: 1,
 				ProtoMinor: 1,
 				Header: http.Header{
@@ -407,7 +406,7 @@ func TestDialerHandshake(t *testing.T) {
 		{
 			name: "bad accept",
 			res: &http.Response{
-				StatusCode: 101,
+				StatusCode: http.StatusSwitchingProtocols,
 				ProtoMajor: 1,
 				ProtoMinor: 1,
 				Header: http.Header{
@@ -421,7 +420,7 @@ func TestDialerHandshake(t *testing.T) {
 		{
 			name: "bad subprotocol",
 			res: &http.Response{
-				StatusCode: 101,
+				StatusCode: http.StatusSwitchingProtocols,
 				ProtoMajor: 1,
 				ProtoMinor: 1,
 				Header: http.Header{
@@ -436,7 +435,7 @@ func TestDialerHandshake(t *testing.T) {
 		{
 			name: "bad extensions",
 			res: &http.Response{
-				StatusCode: 101,
+				StatusCode: http.StatusSwitchingProtocols,
 				ProtoMajor: 1,
 				ProtoMinor: 1,
 				Header: http.Header{
@@ -457,7 +456,7 @@ func TestDialerHandshake(t *testing.T) {
 				// set in test case.
 				req, err := http.ReadRequest(bufio.NewReader(client))
 				if err != nil {
-					t.Fatal(err)
+					panic(err)
 				}
 
 				switch test.accept {
@@ -479,7 +478,7 @@ func TestDialerHandshake(t *testing.T) {
 				var buf bytes.Buffer
 				for _, f := range test.frames {
 					if err := WriteFrame(&buf, f); err != nil {
-						t.Fatal(err)
+						panic(err)
 					}
 					bts = append(bts, buf.Bytes()...)
 					buf.Reset()
@@ -611,7 +610,7 @@ func TestDialerCancelation(t *testing.T) {
 					if t.IsZero() {
 						return nil
 					}
-					d := t.Sub(time.Now())
+					d := time.Until(t)
 					if d < 0 {
 						deadline <- ioErrDeadline
 					} else {
@@ -691,7 +690,7 @@ func BenchmarkDialer(b *testing.B) {
 		rand.Seed(0)
 
 		resp := &http.Response{
-			StatusCode: 101,
+			StatusCode: http.StatusSwitchingProtocols,
 			ProtoMajor: 1,
 			ProtoMinor: 1,
 			Header: http.Header{
@@ -827,18 +826,21 @@ func (s stubConn) Close() error {
 	}
 	return nil
 }
+
 func (s stubConn) SetDeadline(t time.Time) error {
 	if s.setDeadline != nil {
 		return s.setDeadline(t)
 	}
 	return nil
 }
+
 func (s stubConn) SetReadDeadline(t time.Time) error {
 	if s.setReadDeadline != nil {
 		return s.setReadDeadline(t)
 	}
 	return nil
 }
+
 func (s stubConn) SetWriteDeadline(t time.Time) error {
 	if s.setWriteDeadline != nil {
 		return s.setWriteDeadline(t)
